using System;
using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.Threading;
using System.Text.RegularExpressions;

namespace Cotyounoyume{
	public class CollisionFailureDetector
	{
		public EBATCore core;
		public float failureCheckThresholdTime = 7.0f;
		public Dictionary<string, CollisionTrigger> triggerMap;
		public Dictionary<string, bool> checkOnMap;
		public Dictionary<string, bool> failureMap;
		public Dictionary<string, bool> alwaysOnMap;
		public Dictionary<string, float> lastDeactiveTimeMap;
		public Dictionary<string, float> lastActiveTimeMap;
		public string[] cTriggerNames = {"MouthTrigger", "ThroatTrigger","rNippleTrigger", "lNippleTrigger", "LabiaTrigger", "VaginaTrigger", "DeepVaginaTrigger", "DeeperVaginaTrigger"};
		public Dictionary<string, string> zoneMap = new Dictionary<string, string>(){
			{"MouthTrigger", "Mouth"}, 
			{"ThroatTrigger", "Throat"},
			{"rNippleTrigger", "Nipple"}, 
			{"lNippleTrigger", "Nipple"},
			{"LabiaTrigger", "Labia"},
			{"VaginaTrigger", "Vagina"},
			{"DeepVaginaTrigger", "VaginaDeep"},
			{"DeeperVaginaTrigger", "VaginaDeeper"},			
		};
		
		public Dictionary<string, string> zoneMapInv = new Dictionary<string, string>(){
			{"Mouth"       , "MouthTrigger"       }, 
			{"Throat"      , "ThroatTrigger"      },
			{"Nipple"      , "NippleTrigger"      }, 
			{"Labia"       , "LabiaTrigger"       },
			{"Vagina"      , "VaginaTrigger"      },
			{"VaginaDeep"  , "DeepVaginaTrigger"  },
			{"VaginaDeeper", "DeeperVaginaTrigger"},			
		};
		
		public float[] cuaStageThreshold = {
//			3.2f, 3.0f, 2.8f, 2.5f,
			1.2f, 1.0f, 0.8f, 0.5f,
		};
		
		public CollisionFailureDetector(EBATCore core){
			this.core = core;
			triggerMap = new Dictionary<string, CollisionTrigger>();
			checkOnMap = new Dictionary<string, bool>();
			failureMap = new Dictionary<string, bool>();
			alwaysOnMap = new Dictionary<string, bool>();
			lastActiveTimeMap = new Dictionary<string, float>();
			lastDeactiveTimeMap = new Dictionary<string, float>();
			
			Init();
		}
		
		public void Init(){
			OnDestroy();
			
			foreach(string cTriggerName in cTriggerNames){
				triggerMap.Add(cTriggerName, core.atom.GetStorableByID(cTriggerName) as CollisionTrigger);
				checkOnMap.Add(cTriggerName, false);
				failureMap.Add(cTriggerName, false);
				alwaysOnMap.Add(cTriggerName, false);
				lastActiveTimeMap.Add(cTriggerName, 0f);
				lastDeactiveTimeMap.Add(cTriggerName, 0f);
			}
		}
		
		public void OnDestroy(){
			triggerMap.Clear();
			checkOnMap.Clear();
			failureMap.Clear();
			lastActiveTimeMap.Clear();
		}

		public CollisionTrigger GetTrigger(string name){
			if(!triggerMap.ContainsKey(name)){
				return null;
			}
			return triggerMap[name];
		}
		
		public void ActivateTrigger(string name){
			if(triggerMap.ContainsKey(name)){
				core.StartCoroutine(ActivateTriggerCoroutine(name));
			}
		}
		
		public IEnumerator ActivateTriggerCoroutine(string name){
			CollisionTrigger cTrigger = triggerMap[name];
			
			cTrigger.trigger.active = true;
			yield return new WaitForSeconds(0.1f);
			cTrigger.trigger.active = false;
			yield return null;
		}
		
		public void Subscribe(string name){
			if(!triggerMap.ContainsKey(name)){
				core.Info("ERROR: unknown trigger: " + name);
				return;
			}
			
			checkOnMap[name] = true;
			failureMap[name] = false;
			alwaysOnMap[name] = false;
			lastActiveTimeMap[name] = core.currentTime + 5f;
			lastDeactiveTimeMap[name] = core.currentTime + 5f;
		}
		
		public void Unsubscribe(string name){
			if(!triggerMap.ContainsKey(name)){
				return;
			}
			
			checkOnMap[name] = false;
			failureMap[name] = false;
			alwaysOnMap[name] = false;
			lastActiveTimeMap[name] = 0f;
			lastDeactiveTimeMap[name] = 0f;
		}
		
		/*
		public void registerActiveResult(string name){
			try{
			string triggerName = zoneMapInv[name];
			if(triggerName == "NippleTrigger"){
				failureMap["r" + triggerName] = false;
				alwaysOnMap["r" + triggerName] = true;
				lastActiveTimeMap["r" + triggerName] = core.currentTime;				
				
				failureMap["l" + triggerName] = false;
				alwaysOnMap["l" + triggerName] = true;
				lastActiveTimeMap["l" + triggerName] = core.currentTime;				
			}else{
				failureMap[triggerName] = false;
				alwaysOnMap[triggerName] = true;
				lastActiveTimeMap[triggerName] = core.currentTime;				
			}
				
			}catch(Exception e){
				core.loggerE(e);
			}
		}
		*/
		
		public void Update(){
//			debugPrint();
			CheckCollisions();
			RingFailureCollisions();
		}
		
		public void CheckCollisions(){
			foreach(var name in cTriggerNames){
				CheckCollision(name);
			}
		}

		public void CheckCollision(string name){
			if(!checkOnMap[name]){ return; }
			if(triggerMap[name].trigger.active){
				lastActiveTimeMap[name] = core.currentTime;
				if(core.currentTime > lastDeactiveTimeMap[name] + cuaStageThreshold[core.GetCUAStage()]){
					alwaysOnMap[name] = true;
//					core.Info("alwaysOnMap: lastCheckTime:" + lastActiveTimeMap[name]);
//					core.Info("alwaysOnMap: lastDeactiveTimeMap:" + lastDeactiveTimeMap[name]);
//					core.Info("alwaysOnMap: cuaStageThreshold:" + cuaStageThreshold[core.GetCUAStage()]);
				}
			}else{
				lastDeactiveTimeMap[name] = core.currentTime;
//				if(core.currentTime > lastActiveTimeMap[name] + cuaStageThreshold[core.GetCUAStage()]){
//					failureMap[name] = true;
//					lastActiveTimeMap[name] = core.currentTime;
//					core.Info("failureMap: lastCheckTime:" + lastActiveTimeMap[name]);
//					core.Info("failureMap: lastDeactiveTimeMap:" + lastDeactiveTimeMap[name]);
//					core.Info("failureMap: cuaStageThreshold:" + cuaStageThreshold[core.GetCUAStage()]);
//				}
			}
		}
		
		public void RingFailureCollisions(){
//			return;
			foreach(var name in cTriggerNames){
				/*
				if(checkOnMap[name] && failureMap[name]){
					core.OnErogenousZoneTouch(zoneMap[name]);
					failureMap[name] = false;
				}
				*/
				if(checkOnMap[name] && alwaysOnMap[name]){
					triggerMap[name].trigger.active = false;
					lastDeactiveTimeMap[name] = core.currentTime;
					alwaysOnMap[name] = false;
				}
			}
		}
		
		public void DebugPrint(){
			foreach(var name in cTriggerNames){
				core.Info("checkOnMap: " + name + ", " + checkOnMap[name]);
				core.Info("failureMap: " + name + ", " + failureMap[name]);
				core.Info("alwaysOnMap: " + name + ", " + alwaysOnMap[name]);
			}
		}
		
	}
}
