#region using
using System;
using UnityEngine;
using UnityEngine.UI;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.Threading;
using System.Text.RegularExpressions;
using MeshVR;
using GPUTools.Cloth.Scripts;
using GPUTools.Cloth.Scripts.Runtime.Data;
using GPUTools.Cloth.Scripts.Geometry.Data;
#endregion

namespace Cotyounoyume{
	public class ClothesStateChanger
	{
		public List<ClothManager> clothManagers = new List<ClothManager>();
		public List<ClothManager> clothManagersUndressed = new List<ClothManager>();
		public Atom atom;
		public EBATCore core;
		public string[] ngwords;
		public bool holdNGWords   = true;
		public bool holdClothSim  = true;
		public bool holdSkinWraps = true;
		public bool wrapAlterToSimOn = false;
		public bool doneOnce = false;
		
		public ClothesStateChanger(Atom atom, EBATCore core, string[] ngwords){
			this.core = core;
			
			clothManagers = new List<ClothManager>();
			clothManagersUndressed = new List<ClothManager>();
			this.ngwords = ngwords;
			this.atom = atom;
			ChangeClothes();
			doneOnce = true;
		}

		public void ReloadStatus(){
			foreach(var cm in clothManagers){
				cm.ReloadStatus();
			}
			foreach(var cm in clothManagersUndressed){
				cm.ReloadStatus();
			}
		}
		
		public Dictionary<DAZClothingItem, int> MakeHitMap(){
			try{
				JSONStorable geometryStorable = atom.GetStorableByID("geometry");
				DAZCharacterSelector geometryForClothes = geometryStorable as DAZCharacterSelector;			
				List<DAZClothingItem> newClothes = geometryForClothes.clothingItems.ToList();
				
				Dictionary<DAZClothingItem, int> hitMap = new Dictionary<DAZClothingItem, int>();
				foreach(var cm in clothManagers){
					if(!hitMap.ContainsKey(cm.cloth)){
						hitMap.Add(cm.cloth, 1);
					}else{
						hitMap[cm.cloth] += 1;
					}
				}
				foreach(var dci in newClothes){
					if(!hitMap.ContainsKey(dci)){
						hitMap.Add(dci, -1);
					}else{
						hitMap[dci] -= 1;
					}
				}
				return hitMap;
			}catch(Exception e){
				Utility.LoggerE(e);
				return null;
			}
		}
		
		public void ChangeClothesIfNeeded(){
			ReloadStatus();
			bool isNeeded = false;
			Dictionary<DAZClothingItem, int> hitMap = MakeHitMap();
			foreach(KeyValuePair<DAZClothingItem, int> kvp in hitMap){
				if(kvp.Value != 0){
					isNeeded = true;
				}
			}
			if(isNeeded){
				ChangeClothes();
			}
		}
		
		public void ChangeClothes(){
			try{
				JSONStorable geometryStorable = atom.GetStorableByID("geometry");
				DAZCharacterSelector geometryForClothes = geometryStorable as DAZCharacterSelector;
				List<DAZClothingItem> clothes = geometryForClothes.clothingItems.ToList();
				
				Dictionary<DAZClothingItem, int> hitMap = MakeHitMap();				
				foreach(KeyValuePair<DAZClothingItem, int> kvp in hitMap){
					if(kvp.Value == 0){			//exists both array
						//nop
					}else if(kvp.Value == 1){	//exists only clothManagers
						RecoverSpecificCloth(kvp.Key);
						RemoveSpecificClothFromManager(kvp.Key);
					}else if(kvp.Value == -1){	//exists only new clothes
						if(kvp.Key.active && kvp.Key.creatorName != "cotyounoyume"){
							clothManagers.Add(new ClothManager(core, kvp.Key));
						}
					}
				}
			}catch(Exception e){
				SuperController.LogMessage("*** Exception:" + e);
			}			
		}
		
		public void SetRemoveOptions(bool bHoldNGWords, bool bHoldClothSim, bool bHoldSkinWraps, bool wrapAlterToSimOn){
			this.holdNGWords      = bHoldNGWords;
			this.holdClothSim     = bHoldClothSim;
			this.holdSkinWraps    = bHoldSkinWraps;
			this.wrapAlterToSimOn = wrapAlterToSimOn;
			SetAlterOption();
		}
		
		public void SetAlterOption(){
			foreach(var cm in clothManagers){
				cm.SetAlterOption(wrapAlterToSimOn);
			}
		}
		
		public bool IsNG(DAZClothingItem dci){
			bool result = false;
			foreach(var name in ngwords){
				if(dci.displayName.ToLower().Contains(name)){
					result = true;
				}
			}
			return result;
		}
		
		public void RemoveClothes(){
			try{
				
			int clothNum = clothManagers.Count;
			for(int i = clothNum - 1; i >= 0; i--){
				bool hold = false;
				bool hasRuntime = clothManagers[i].hasRuntime;
				bool hasCs      = clothManagers[i].hasCs;
				bool hasAltWrap = clothManagers[i].hasAltWrap;
				DAZClothingItem cloth = clothManagers[i].cloth;

				if(holdNGWords){
					if(IsNG(cloth)){
						hold = true;
					}
				}
				if(holdClothSim){
					if(hasRuntime){
						hold = true;
					}
				}
				if(holdSkinWraps){
					if(hasAltWrap){
						hold = true;
					}
				}
				if(!hold){
					RemoveClothAndBackup(i);
				}
			}
			
			}catch(Exception e){
				Utility.LoggerE(e);
			}
		}

		public void RemoveClothAndBackup(int i){
			ClothManager cm = clothManagers[i];
			clothManagersUndressed.Add(cm);
			JSONStorable geometryStorable = atom.GetStorableByID("geometry");
			DAZCharacterSelector geometryForClothes = geometryStorable as DAZCharacterSelector;
			geometryForClothes.SetActiveClothingItem(cm.cloth, active: false);
			clothManagers.RemoveAt(i);
		}
		
		public void RestoreClothFromBackup(int i){
			ClothManager cm = clothManagersUndressed[i];
			clothManagers.Add(cm);
			JSONStorable geometryStorable = atom.GetStorableByID("geometry");
			DAZCharacterSelector geometryForClothes = geometryStorable as DAZCharacterSelector;
			geometryForClothes.SetActiveClothingItem(cm.cloth, active: true);
			clothManagersUndressed.RemoveAt(i);
		}
		
		public void RemoveResetClothes(){
			int clothNum = clothManagersUndressed.Count;
			for(int i = clothNum - 1; i >= 0; i--){
				RestoreClothFromBackup(i);
			}
		}
		
		public void CheckClothAttribute(){
			ChangeClothesIfNeeded();
			for(int i = 0; i < clothManagers.Count; i++){
				clothManagers[i].CheckClothAttribute();
			}
		}
		
		public void RecoverSpecificCloth(DAZClothingItem dci){
			foreach(var cm in clothManagers){
				if(dci == cm.cloth){
					cm.Recover();
					cm.WrapNormal();
				}
			}
		}
		
		public void RemoveSpecificClothFromManager(DAZClothingItem dci){
			int clothNum = clothManagers.Count;
			for(int i = clothNum - 1; i >= 0; i--){
				if(dci == clothManagers[i].cloth){
					clothManagers.RemoveAt(i);
				}
			}
		}
		
		public void Recover(){
			try{
				for(int i = 0; i < clothManagers.Count; i++){
					clothManagers[i].Recover();
				}
				for(int i = 0; i < clothManagersUndressed.Count; i++){
					clothManagersUndressed[i].Recover();
				}
			}catch(Exception e){
				SuperController.LogMessage("*** Exception:" + e);
			}			
		}
		
		public void Reset(){
			try{
				for(int i = 0; i < clothManagers.Count; i++){
					clothManagers[i].Reset();
				}
				for(int i = 0; i < clothManagersUndressed.Count; i++){
					clothManagersUndressed[i].Reset();
				}
			}catch(Exception e){
				SuperController.LogMessage("*** Exception:" + e);
			}			
		}
		
		public void WrapAlter(bool isForce = false){
			for(int i = 0; i < clothManagers.Count; i++){
				clothManagers[i].wm.WrapAlter();
			}
		}
		
		public void WrapNormal(){
			for(int i = 0; i < clothManagers.Count; i++){
				clothManagers[i].wm.WrapNormal();
			}
		}
		
		public void Change(string type = "UndressStrong", float ratio = 1.0f){
			try{
				for(int i = 0; i < clothManagers.Count; i++){
					if(!holdNGWords || !IsNG(clothManagers[i].cloth)){
						clothManagers[i].Change(type, ratio);
					}
				}
			}catch(Exception e){
				SuperController.LogMessage("*** Exception:" + e);
			}			
		}
		
		public void InitialChange(string type){
			try{
				for(int i = 0; i < clothManagers.Count; i++){
					if(!holdNGWords || !IsNG(clothManagers[i].cloth)){
						clothManagers[i].InitialChange(type);
					}
				}
			}catch(Exception e){
				SuperController.LogMessage("*** Exception:" + e);
			}			
		}
		
		public void UndressStop(){
			try{
				for(int i = 0; i < clothManagers.Count; i++){
					clothManagers[i].UndressStop();
				}
			}catch(Exception e){
				SuperController.LogMessage("*** Exception:" + e);
			}			
		}
		
		public void OnDestroy(){
			if(doneOnce){
				for(int i = 0; i < clothManagers.Count; i++){
					clothManagers[i].OnDestroy();
					clothManagers[i] = null;
				}
				clothManagers = null;
				for(int i = 0; i < clothManagersUndressed.Count; i++){
					clothManagersUndressed[i].OnDestroy();
					clothManagersUndressed[i] = null;
				}
				clothManagers = null;
			}
		}
	}
	
	public class ClothManager
	{
		public EBATCore core;
		public DAZClothingItem cloth;
		public WrapManager wm;
		public ClothSettingsManager csm;
		
		public bool simEnabledOrig = false;
		public bool isBuiltin = false;
		public bool hasRuntime = false;
		public bool hasCsc     = false;
		public bool hasCs      = false;
		public bool hasAltWrap = false;
		public bool doneOnce = false;
		public bool wrapAlterToSimOn = false;
		
		public ClothManager(EBATCore core, DAZClothingItem cloth){
			this.core = core;
			this.cloth = cloth;
			
			wm  = new WrapManager(core, cloth);
			csm = new ClothSettingsManager(core, cloth);
			
			isBuiltin = !cloth.uid.Contains("/");
			hasCsc     = csm.hasCsc;
			hasCs      = csm.hasCs;
			hasRuntime = wm.hasRuntime;
			hasAltWrap = wm.hasAltWrap;
			doneOnce = true;
		}
		
		public void ReloadStatus(){
			csm.ReloadStatus();
			wm.ReloadStatus();
			
			hasCsc     = csm.hasCsc;
			hasCs      = csm.hasCs;
			hasRuntime = wm.hasRuntime;
			hasAltWrap = wm.hasAltWrap;
		}
		
		public bool IsIrregular(){
			return csm.isIrregular;
		}
		
		public void SetAlterOption(bool wrapAlterToSimOn){
			wm.SetAlterOption(wrapAlterToSimOn);
		}
		
		public void CheckClothAttribute(){
			try{
				Utility.Info("clothname:" + cloth.displayName);
				bool hasCsc = (csm.csc != null);
				bool hasCs  = (csm.cs  != null);
				bool hasGeometryData = false;
				ClothGeometryData geometryData = null;
				if(hasCs){
					geometryData = csm.cs.GeometryData;
					hasGeometryData = (geometryData != null);
				}
				if(hasCs != hasRuntime){
					Utility.Info(" ******** IRREGULAR DATA *********:" +  cloth.displayName + "(" + cloth.uid + ")");
					if(hasCs){
						csm.cs.ProcessGeometry();
					}
				}
				Utility.Info("  hasGeometryData:" +  hasGeometryData);
				Utility.Info("  uid:" +  cloth.uid);
				Utility.Info("  hasAltWrap:" +  hasAltWrap);
				Utility.Info("  hasRuntime:" +  hasRuntime);
				Utility.Info("  wm:" +  wm);
				Utility.Info("  csm:" +  csm);
				Utility.Info("  hasCsc:" +  hasCsc);
				Utility.Info("  hasCs:" +  hasCs);
			}catch(Exception e){
				Utility.LoggerE(e);
			}
			
		}
		
		public void WrapNormal(){
			wm.Recover();
		}
		public void Recover(){
			csm.Recover();
		}
		
		public void Reset(){
			csm.Reset();
		}
		
		public void Change(string type, float ratio){
			csm.Change(type, ratio);
		}
		
		public void InitialChange(string type){
			csm.InitialChange(type);
		}
		
		public void UndressStop(){
			csm.UndressStop();
		}
		
		public void OnDestroy(){
			if(doneOnce){
				wm.OnDestroy();
				csm.OnDestroy();
			}
		}
	}
	
	public class ClothSettingsManager
	{
		public EBATCore core;
		public ClothSimControl csc;
		public ClothSettings   cs;
		public RuntimeData     rtd;
		public DAZClothingItem cloth;
		
		public bool hasRuntime = false;
		public bool hasCsc     = false;
		public bool hasCs      = false;
		public bool isIrregular = false;
		
		public bool    backupBreakEnabled;
		public float   backupBreakThreshold;
		public float   backupFriction;
		public float   backupWeight;
		public float   backupGravityMultiplier;
		public float   backupStiffness;
		public float   backupDistanceScale;
		public bool    backupSimEnabled;
		public float   backupCollisionPower;
		public float   backupParticleRadius;
		public float   backupStretchability;
		public float   backupJointStrength;
		
		public float[] backupParticlesStrength;
		public float[] backupParticlesBlend;
		public bool    doneOnce = false;
		public bool    arrayCopyDone = false;
		public bool    csCopyDone = false;

		public ClothSettingsManager(EBATCore core, DAZClothingItem cloth){
			this.core = core;
			this.cloth = cloth;
			ReloadStatus();
			doneOnce = true;
		}	

		public void ReloadStatus(){
			if(!csCopyDone || !arrayCopyDone){
				csc = Utility.GetClothSimControl(cloth);
				cs  = (csc != null) ? Utility.GetClothSettings(csc) : null ;
				rtd = Utility.GetRuntime(cloth);
				hasRuntime = (rtd != null);
				hasCsc     = (csc != null);
				hasCs      = (cs  != null);
				
				if(hasCs && csCopyDone == false){
					backupBreakEnabled      = cs.BreakEnabled;
					backupBreakThreshold    = cs.BreakThreshold;
					backupFriction          = cs.Friction;
					backupWeight            = cs.Weight;
					backupGravityMultiplier = cs.GravityMultiplier;
					backupStiffness         = cs.Stiffness;
					backupDistanceScale     = cs.DistanceScale;
					backupCollisionPower    = cs.CollisionPower;
					backupParticleRadius    = cs.ParticleRadius;
					backupStretchability    = cs.Stretchability;
					backupJointStrength     = cs.JointStrength;
					
					backupSimEnabled        = csc.simEnabledJSON.val;
					csCopyDone = true;
				}
				
				if(hasRuntime && arrayCopyDone == false){
					backupParticlesStrength = new float[cs.GeometryData.ParticlesStrength.Length];
					backupParticlesBlend    = new float[cs.GeometryData.ParticlesBlend.Length];
					Array.Copy(cs.GeometryData.ParticlesStrength, backupParticlesStrength, cs.GeometryData.ParticlesStrength.Length);
					Array.Copy(cs.GeometryData.ParticlesBlend   , backupParticlesBlend   , cs.GeometryData.ParticlesBlend.Length   );
					arrayCopyDone = true;
				}
			}
		}
		
		public IEnumerator SimEnableCoroutine(){
			if(!csc.simEnabledJSON.val){
				csc.SetSimEnabled(true);
				yield return new WaitForSeconds(0.1f);
				csc.SetSimEnabled(false);
				csc.Reset();
			}
			yield return null;
		}
		
		public void Recover(){
			if(!hasRuntime){ return; }
			
			csc.clothSettings.BreakEnabled      = backupBreakEnabled;
			csc.clothSettings.BreakThreshold    = backupBreakThreshold;
			csc.clothSettings.Friction          = backupFriction;
			csc.clothSettings.Weight            = backupWeight;		
			csc.clothSettings.GravityMultiplier = backupGravityMultiplier;
			csc.clothSettings.Stiffness         = backupStiffness;
			csc.clothSettings.DistanceScale     = backupDistanceScale;
			csc.clothSettings.CollisionPower    = backupCollisionPower;
			csc.clothSettings.ParticleRadius    = backupParticleRadius;
			csc.clothSettings.Stretchability    = backupStretchability;
			csc.clothSettings.JointStrength     = backupJointStrength;

			csc.SetSimEnabled(backupSimEnabled);
			if(hasRuntime){
				Array.Copy(backupParticlesStrength, cs.GeometryData.ParticlesStrength, backupParticlesStrength.Length);
				Array.Copy(backupParticlesBlend   , cs.GeometryData.ParticlesBlend   , backupParticlesBlend.Length);
			}
		}
		
		public void Reset(){
			if(!hasRuntime){ return; }
			
			csc.Reset();
		}

		public void UndressStop(){
			if(!hasRuntime){ return; }
			
			csc.clothSettings.Friction          = 1.0f;
			csc.clothSettings.Weight            = 0.5f;
			csc.clothSettings.GravityMultiplier = 1.0f;
			csc.clothSettings.Stiffness = 1f;
			csc.clothSettings.DistanceScale = 1f;
			csc.clothSettings.BreakEnabled   = false;
			csc.clothSettings.BreakThreshold = 1f;
		}
		
		public void SettingForUndress(){
			csc.SetSimEnabled(true);
			csc.clothSettings.BreakEnabled   = true;
			csc.clothSettings.BreakThreshold = 0f;					
			csc.clothSettings.JointStrength = 0.05f;			
		}
		
		public void SettingForForce(){
			cs.DistanceScale  = 1.01f;
			cs.JointStrength = 0.05f;
			cs.CollisionPower = 0.5f;
			if(cs.GeometryData != null){
				for(int i = 0; i < cs.GeometryData.ParticlesStrength.Length; i++){
//					cs.GeometryData.ParticlesStrength[i] = 0.05f;
					cs.GeometryData.ParticlesBlend[i] = 0.05f;
				}
			}
			Reset();				
		}
		
		public void InitialChange(string type){
			if(!hasCs){ return; }
//			if(!hasRuntime){ return; }
			
			if(type == "Undress"){
				SettingForUndress();
			}else if(type == "UndressMiddle"){
				SettingForUndress();
				SettingForForce();
			}else if(type == "UndressForce"){
				SettingForUndress();
				SettingForForce();
			}else if(type == "Break"){
				csc.SetSimEnabled(true);
				csc.clothSettings.BreakEnabled   = true;
				csc.clothSettings.BreakThreshold = 0.1f;
				csc.SyncAllowDetatch(true);
			}else if(type == "Break2"){
				SettingForUndress();
				csc.SyncAllowDetatch(true);
			}else if(type == "Reverse"){
				csc.clothSettings.Stiffness = 1f;
			}else if(type == "UndressAndReverse"){
				SettingForUndress();
				csc.clothSettings.Stiffness = 1f;
			}else if(type == "UndressAndReverseForce"){
				SettingForUndress();
				csc.clothSettings.Stiffness = 1f;
				SettingForForce();
			}else{
				Utility.Info("In ClothSimControl.Change: Unknown Command: " + type);
			}
		}
		
		public void Change(string type, float ratio){
			if(!hasCs){ return; }
//			if(!hasRuntime){ return; }
			
			if(type == "Undress"){
				Undress(ratio);
			}else if(type == "UndressMiddle"){
				Undress(ratio);
			}else if(type == "UndressForce"){
				Undress(ratio, isForce:true);
			}else if(type == "Break"){
			}else if(type == "Break2"){
			}else if(type == "Reverse"){
				Reverse(ratio);
			}else if(type == "UndressAndReverse"){
				Reverse(ratio);
			}else if(type == "UndressAndReverseForce"){
				Reverse(ratio, isForce:true);
			}else{
				Utility.Info("In ClothSimControl.Change: Unknown Command: " + type);
			}
		}
		
		public void Undress(float ratio, bool isForce = false){
			if(!hasCs){ return; }
//			if(!hasRuntime){ return; }
			
			float multi = (isForce) ? 2.0f : 1.0f;
			csc.clothSettings.Friction          = Mathf.Lerp(1.0f, 0.1f, ratio * multi);
			csc.clothSettings.Weight            = Mathf.Lerp(0.5f, 2.0f, ratio * multi);
			csc.clothSettings.GravityMultiplier = Mathf.Lerp(0.5f, 2.0f, ratio * multi);			
		}
		
		public void Reverse(float ratio, bool isForce = false){
			if(!hasCs){ return; }
//			if(!hasRuntime){ return; }
			
			float multi = (isForce) ? 2.0f : 1.0f;
			csc.clothSettings.Friction          = Mathf.Lerp(1.0f,  0.1f, ratio * multi);
			csc.clothSettings.Weight            = Mathf.Lerp(0.5f,  2.0f, ratio * multi);
			csc.clothSettings.GravityMultiplier = Mathf.Lerp(1.0f, -2.0f, ratio * multi);
		}
				
		public void OnDestroy(){
			if(!hasCs){ return; }
//			if(!hasRuntime){ return; }
			
			if(doneOnce){
				Recover();
				
				csc = null;
				cs  = null;
				rtd = null;
				backupParticlesStrength = null;
				backupParticlesBlend = null;
			}else{
				//nop
			}
		}
	}

	public class WrapManager
	{
		public EBATCore core;
		public DAZClothingItem cloth;
		public bool hasAltWrap = false;
		public bool hasRuntime = false;
		public List<string> AltWrapNames;
		public DAZSkinWrapSwitcher wrapSwitcher = null;
		public bool doneOnce = false;
		public bool wrapAlterToSimOn = false;
		
		public WrapManager(EBATCore core, DAZClothingItem cloth){
			this.core = core;
			this.cloth = cloth;
			ReloadStatus();
			doneOnce = true;
		}
		
		public void ReloadStatus(){
			if(!hasAltWrap){
				AltWrapNames = new List<string>();
				hasRuntime = Utility.HasRuntime(cloth);
				
				wrapSwitcher = this.cloth.GetComponentInChildren<DAZSkinWrapSwitcher>();
				if(wrapSwitcher != null && wrapSwitcher.currentWrapName != ""){
					hasAltWrap = true;
					if(wrapSwitcher.IsStringChooserJSONParam("wrapName")){
						foreach(var name in wrapSwitcher.GetStringChooserJSONParamChoices("wrapName")){
							if(name != "Normal"){
								AltWrapNames.Add(name);
							}
						}
					}
				}else{
					wrapSwitcher = null;
				}
			}
		}
		
		public void SetAlterOption(bool wrapAlterToSimOn){
			this.wrapAlterToSimOn = wrapAlterToSimOn;
		}
		
		public void Recover(){
			if(!hasAltWrap){ return; }
			
			WrapNormal();
		}
		
		public void WrapNormal(){
			if(!hasAltWrap){ return; }
			
			if(hasAltWrap && wrapSwitcher.currentWrapName != "Normal"){
				wrapSwitcher.SetCurrentWrapName("Normal");
			}
		}
		
		public void WrapAlter(){
			if(!hasAltWrap){ return; }
			
			if(hasAltWrap){
				if(wrapAlterToSimOn || !hasRuntime){
					int min = 0;
					int max = AltWrapNames.Count;
					int rand = UnityEngine.Random.Range(min, max);
					string altWrapName = AltWrapNames[rand];
					wrapSwitcher.SetCurrentWrapName(altWrapName);
				}
			}
		}
		
		public void OnDestroy(){
			if(!hasAltWrap){ return; }
			
			if(doneOnce){
				WrapNormal();
				AltWrapNames.Clear();
				AltWrapNames = null;
				wrapSwitcher = null;
			}else{
				//nop
			}
		}
	}
}